#pragma once

/*
Copyright © Glensound Electronics Ltd and individual contributors. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met: Redistributions of source code must retain the
above copyright notice, this list of conditions and the following disclaimer. Redistributions
in binary form must reproduce the above copyright notice, this list of conditions and the
following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of Glensound nor the names of its contributors may be used to endorse or
promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Content based on github.com/github/opensource.guide used under the CC-BY-4.0 license.
*/

#include "GsProtocol.h"


namespace Gs
{

	// Command code system.
	// The two MSb of the command byte indicate the control message length category.
	// Variable-length messages use a BlockHeader containing the message size.
	//
	// code			group		Rel. range	Abs. range
	// -----------------------------------------------
	// 00000000		reserved
	// 0xxxxxxx		4 bytes		1 to 127	1 to 127
	// 10xxxxxx		8 bytes		0 to 63		128 to 191
	// 11xxxxxx		variable	0 to 63		192 to 255

	constexpr uint8_t cFirstCmd4 = 1;
	constexpr uint8_t cFirstCmd8 = 128;
	constexpr uint8_t cFirstCmdVar = 192;

	
	// Standard commands are either supported or reserved by all device types.
	// Documentation of each particular device type states which of the standard
	// commands are supported.
	//
	enum class Command : uint8_t
	{
		// 4-byte commands
		//
		Srate = cFirstCmd4,	// message is Value8
		ClockSource,		// message is Value8
		AdcSettings,		// message is Value8
		DacSettings,		// message is Value8
		_end4,

		// 8-byte commands
		//
		RestoreDefaults = cFirstCmd8,	// message is Gs::RestoreDefaults
		_end8,

		// Variable-length commands
		//
		ChannelName = cFirstCmdVar,		// message is Gs::ChannelName
		_endVar
	};

	
	// Device-specific 4-byte commands start at this value
	//
	constexpr auto cNext4byteCommand = (uint8_t)Command::_end4;
	
	// Device-specific 8-byte commands start at this value
	//
	constexpr auto cNext8byteCommand = (uint8_t)Command::_end8;
	
	// Device-specific variable-length commands start at this value
	//
	constexpr auto cNextVarLengthCommand = (uint8_t)Command::_endVar;
	

	// Generic control message with 8-bit values
	//
	struct alignas(uint32_t) Value8
	{
		uint8_t command;
		uint8_t value;
		uint8_t param1;
		uint8_t param2;
	};
	static_assert(sizeof(Value8) == 4);

	
	// Command::RestoreDefaults
	// Restores factory default settings of selected components
	//
	struct alignas(uint32_t) RestoreDefaults
	{
		Command		command;		// Command::RestoreDefaults
		uint8_t		reserved1;
		uint16_t	reserved2;
		uint32_t	settingsParts;	// a combination of SettingsPart bits
	};
	static_assert(sizeof(RestoreDefaults) == 8);

	
	// Command::ChannelName
	// Sets the name of a Dante channel
	//
	struct ChannelName
	{
		union
		{
			BlockHeader		header;
			struct
			{
				Command			command;	// Command::ChannelName
				DanteChannelId	channelId;
			};
		};
		
		// The payload is the text, in UTF-8 encoding.
		// Text size must be a multiple of 4 bytes, zero-padded if necessary.
	};
	static_assert(sizeof(ChannelName) == 4);
}
