#pragma once

/*
Copyright © Glensound Electronics Ltd and individual contributors. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met: Redistributions of source code must retain the
above copyright notice, this list of conditions and the following disclaimer. Redistributions
in binary form must reproduce the above copyright notice, this list of conditions and the
following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of Glensound nor the names of its contributors may be used to endorse or
promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Content based on github.com/github/opensource.guide used under the CC-BY-4.0 license.
*/

#include "GsTypes.h"

#ifndef __BYTE_ORDER__
#define __ORDER_LITTLE_ENDIAN__ 3412
#define __ORDER_BIG_ENDIAN__ 1234
#define __BYTE_ORDER__ __ORDER_LITTLE_ENDIAN__
#endif


// Examples of data transactions
// =============================
//
// Simple instruction (GetStatus)
// ------------------------------
//
// The host sends a single UDP datagram consisting of Header only:
//
// Header header
// {
//		.magic = cMagic,
//		.size = sizeof(Header),			// no payload, just the header
//		.multipacket = 0,				// one-packet transaction
//		.opcode = Opcode::GetStatus,
//		.flags =
//		{
//			.exclusiveAccess = 0,		// shared mode (or exclusive)
//			.noMeter = 0,				// fast meter updates required
//		},
//		.controllerId = myUniqueId		// a unique number on this network
// };
//
// Simple response (Status)
// ------------------------
//
// The device sends a single UDP datagram consisting of Header followed by status data:
//
// struct MyStatusMessage
// {
//		Header header;
//		Status status;
// } message
// {
//		.header =
//		{
//			.magic = cMagic,
//			.size = sizeof(MyStatusMessage),	// header plus payload
//			.multipacket = 0,					// one-packet transaction
//			.opcode = Opcode::Status,
//			.flags =
//			{
//				.passwordValid = true,
//				.accessGranted = true
//			},
//			.sequence = ++mySequence			// increment before sending
//		},
//		.status =
//		{
//			.flags =
//			{
//				.identify = 0,
//				.clockLocked = 1
//			},
//			.generations[0] = w,	// pseudocode!
//			.generations[1] = x,
//			.generations[2] = y,
//			.generations[3] = z,
//		}
// };
//
// Simple command
// --------------
//
// struct MyCommand
// {
//		Header header;
//		MyCommand command;
// } command
// {
//		.header =
//		{
//			.magic = cMagic,
//			.size = sizeof(Header),			// no payload, just the header
//			.multipacket = 0,				// one-packet transaction
//			.opcode = Opcode::SetControl,
//			.flags =
//			{
//				.exclusiveAccess = 0,		// shared mode (or exclusive)
//				.noMeter = 0,				// fast meter updates required
//			},
//			.controllerId = myUniqueId		// a unique number on this network
//		},
//		.command =
//		{
//			.opcode = Command::SetParam1,
//			.value = 10
//		}
// };


namespace Gs
{
#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
	constexpr uint64_t cMagic = 0x006c727443205347;	// "GS Ctrl\0"
#else
	constexpr uint64_t cMagic = 0x4753204374726c00;	// "GS Ctrl\0"
#endif

	// Maximum size of a UDP packet data payload for a Broadway or Brooklyn (BHIP) device.
	//
	constexpr size_t cMaxBhipPacketSize = 1448;

	// Maximum size of a UDP packet data payload for an Ultimo (UHIP) device.
	//
	constexpr size_t cMaxUhipPacketSize = 500;
	

	enum class Opcode : uint8_t
	{
		Invalid,
		Status,         // status payload follows (device to controller)
		GetStatus,		// request for status, including metering (controller to device)
		SetControl,     // control payload follows (controller to device)
		Info,			// info payload follows (device to controller)
		GetInfo,		// request for info (controller to device)
		Config,			// config payload follows (device to controller)
		GetConfig,		// request for config (controller to device)
		SetConfig,		// config payload follows (controller to device)
		SetPassword,	// new password payload follows (controller to device)
		Report,			// report payload follows (device to controller)
		GetReport,		// request for report (controller to device)
		SetPreset,		// data block follows (controller to device)
		SetUpdate,		// data block follows (host to device)
		UpdaterReport	// updater progress payload follows (device to host)
	};
	
	
	// Common header for all GS protocol UDP packets
	//
	struct Header
	{
		uint64_t	magic;					// cMagic (for commands, XORed with password)
		struct
		{
			uint16_t	size : 15;			// size of the packet in bytes (including this header)
			uint16_t	multipacket : 1;	// 0 = single-packet message, 1 = multiple-packet message
		} sizeInfo;
		Opcode		opcode;
		union
		{
			struct
			{
				bool exclusiveAccess : 1;	// 0 = shared access, 1 = exclusive access
				bool noMeter : 1;			// 0 = meter data wanted (10 packets per second), 1 = doesn't want meter data
				bool passwordValid : 1;		// 0 = invalid, 1 = password used in GetInfo was valid
				bool accessGranted : 1;		// 0 = not granted, 1 = sender of GetInfo has access granted
			};
			uint8_t value;
		} flags;
		union
		{
			uint32_t	controllerId;		// unique ID of the controller sending this header
			struct
			{
				uint8_t	sequence;			// incremented for each status/config/report response
				uint8_t devId[3];			// (reasonably) unique device ID
			};
		};
	};
	static_assert(sizeof(Header) == 16);
	
	
	// For convenience, here is the same Header when used for controller-to-device transfers:
	//
	struct ControllerHeader
	{
		uint64_t	magic;					// cMagic XORed with password
		struct
		{
			uint16_t	size : 15;			// size of the packet in bytes (including this header)
			uint16_t	multipacket : 1;	// 0 = single-packet message, 1 = multiple-packet message
		} sizeInfo;
		Opcode		opcode;
		union
		{
			struct
			{
				bool exclusiveAccess : 1;	// 0 = shared access request, 1 = exclusive access request
				bool noMeter : 1;			// 0 = want meter data, 1 = doesn't want meter data
			};
			uint8_t value;
		} flags;
		uint32_t	controllerId;			// unique ID of the controller sending this header
	};
	static_assert(sizeof(ControllerHeader) == 16);
	
	
	// For convenience, here is the same Header when used for device-to-controller transfers:
	//
	struct DeviceHeader
	{
		uint64_t	magic;					// cMagic
		struct
		{
			uint16_t	size : 15;			// size of the packet in bytes (including this header)
			uint16_t	multipacket : 1;	// 0 = single-packet message, 1 = multiple-packet message
		} sizeInfo;
		Opcode		opcode;
		union
		{
			struct
			{
				bool exclusiveAccess : 1;	// 0 = shared access mode, 1 = exclusive access mode
				bool reserved : 1;			// ignore
				bool passwordValid : 1;		// 0 = invalid, 1 = password used in GetInfo or GetConfig was valid
				bool accessGranted : 1;		// 0 = not granted, 1 = sender of GetInfo or GetConfig has access granted
			};
			uint8_t value;
		} flags;
		uint8_t	sequence;					// incremented for each status/config/report response
		uint8_t devId[3];					// (reasonably) unique device ID
	};
	static_assert(sizeof(DeviceHeader) == 16);

	
	// This structure directly follows the GS header when the 'multipacket' field is set to 1
	//
	struct MultiPacketInfo
	{
		uint8_t		transactionId;				// sequential transaction number, incremented for every new transaction
		uint8_t		part;						// part number of the current transaction
		uint16_t	payloadLengthWords;			// total transaction payload length in 32-bit words		
	};
	static_assert(sizeof(MultiPacketInfo) == 4, "Size");


	// Maximum number of parts in one multi-packet transaction
	//
	constexpr uint cMaxTransactionParts{ 32 };

	// Maximum data payload sizes for Broadway- or Brooklyn-based devices
	//
	constexpr size_t cMaxBhipSinglePacketPayloadSize = cMaxBhipPacketSize - sizeof(Header);							// 1432
	constexpr size_t cMaxBhipMultiPacketPayloadSize = cMaxBhipSinglePacketPayloadSize - sizeof(MultiPacketInfo);	// 1428
	constexpr size_t cMaxBhipTransactionPayloadSize{ cMaxTransactionParts * cMaxBhipMultiPacketPayloadSize };		// 45696

	// Maximum data payload sizes for Ultimo-based devices
	//
	constexpr size_t cMaxUhipSinglePacketPayloadSize = cMaxUhipPacketSize - sizeof(Header);							// 484
	constexpr size_t cMaxUhipMultiPacketPayloadSize = cMaxUhipSinglePacketPayloadSize - sizeof(MultiPacketInfo);	// 480
	constexpr size_t cMaxUhipTransactionPayloadSize{ cMaxTransactionParts * cMaxUhipMultiPacketPayloadSize };		// 15360

	
	// Variable-length data block header
	//
	struct BlockHeader
	{
		uint8_t		type;						// data/report type or command
		uint8_t		param;						// type-specific, or report generation number
		uint16_t	payloadWords;				// payload length in 32-bit words

		// Payload follows
	};
	static_assert(sizeof(BlockHeader) ==  4, "Size");


	// Response to the GetInfo message (legacy devices only: older models of DARK1616M and DARK8MAI)
	//
	struct LegacyInfo
	{
		uint16_t	firmwareVersion;		// MMMMmmmm ssssssss (M=major, m=minor, s=subminor)
		uint16_t	statusPortBe;			// Big-endian
		uint32_t	statusIpBe;				// Big-endian
		uint16_t	productId;
		uint8_t		variantId;
		uint8_t		protocolVersion;
		uint16_t	reserved;
		uint16_t	danteProcessId;						// part of unique Dante ID (Audinate says "almost always zero")
		uint64_t	danteId;							// part of unique Dante ID (actually, MAC address of the primary IF)
		char		hostName[cMaxDanteNameSize];		// NUL-terminated UTF-8 (e.g. "Paradiso_0af9c0")
		char		friendlyName[cMaxDanteNameSize];	// NUL-terminated UTF-8 (e.g. "Paradiso-Marcin")
		char		domainName[cMaxDanteNameSize];		// NUL-terminated UTF-8
	};
	static_assert(sizeof(LegacyInfo) == 120, "Size");


	// Response to the GetInfo message (all devices except older models of DARK1616M and DARK8MAI)
	//
	struct Info
	{
		uint16_t	firmwareVersion;		// MMMMmmmm ssssssss (M=major, m=minor, s=subminor)
		uint16_t	statusPortBe;			// Big-endian
		uint32_t	statusIpBe;				// Big-endian
		uint16_t	productId;
		uint8_t		variantId;
		uint8_t		protocolVersion;
		struct
		{
			uint16_t dntUpdateable : 1;		// true if DNT file can be updated from Glensound Controller
			uint16_t dfuUpdateable : 1;		// true if the main DFU file can be updated from Glensound Controller
			uint16_t dfu2Updateable : 1;	// true if the secondary DFU file can be updated from Glensound Controller
			uint16_t reserved1 : 13;
		};
		uint16_t	danteProcessId;						// part of unique Dante ID (Audinate says "almost always zero")
		uint64_t	danteId;							// part of unique Dante ID (actually, MAC address of the primary IF)
		char		hostName[cMaxDanteNameSize];		// NUL-terminated UTF-8 (e.g. "Paradiso_0af9c0")
		char		friendlyName[cMaxDanteNameSize];	// NUL-terminated UTF-8 (e.g. "Paradiso-Marcin")
		char		domainName[cMaxDanteNameSize];		// NUL-terminated UTF-8
		uint16_t	dntVersion;							// MMMMmmmm ssssssss (DNT file version)
		uint16_t	dfu2Version;						// MMMMmmmm ssssssss (Secondary DFU file version)
		uint32_t	reserved2;							// reserved - set to zero
	};
	static_assert(sizeof(Info) == 128, "Size");


	// Response to the GetConfig message
	//
	struct Config
	{
		uint32_t	statusIpBe;			// IP address for sending status (big endian)
		uint16_t	statusPortBe;		// port for sending status (big endian)
		uint16_t	reserved;			// reserved
	};
	static_assert(sizeof(Config) == 8, "Size");


	// Payload for Opcode::GetReport
	//
	struct Query
	{
		uint8_t dataType;		// ReportId or a device-specific report ID
		uint8_t reserved1;
		uint16_t reserved2;
	};
	static_assert(sizeof(Query) == 4, "Size");


	// Payload for Opcode::SetPassword
	//
	struct Password
	{
		uint64_t	newPassword;				// new password
		uint64_t	reserved;					// reserved - set to zero
	};
	static_assert(sizeof(Password) == 16, "Size");
}
