#pragma once

/*
Copyright © Glensound Electronics Ltd and individual contributors. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met: Redistributions of source code must retain the
above copyright notice, this list of conditions and the following disclaimer. Redistributions
in binary form must reproduce the above copyright notice, this list of conditions and the
following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of Glensound nor the names of its contributors may be used to endorse or
promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Content based on github.com/github/opensource.guide used under the CC-BY-4.0 license.
*/

#include <cstdint>
#include <cstddef>


namespace Gs
{
	using uint = unsigned int;
	using Volume = uint8_t;
	using Pan = int8_t;
	
	// Maximum size in bytes of a Dante name (channel name, device name, etc.).
	// Dante names use NUL-terminated UTF8 encoding and the size limit applies
	// to the number of bytes including the NUL termination.
	//
	constexpr size_t cMaxDanteNameSize = 32;

	
	enum class Lock : uint8_t
	{
		Unlocked,
		Locked
	};

	enum class ClockSource : uint8_t
	{
		Dante,		// derived from the Dante network
		ExtWck,		// External Word Clock input
		Input,		// one of the digital audio inputs
		Crystal		// local crystal oscillator
	};

	enum class Srate : uint8_t
	{
		srUnknown,
		srAuto = srUnknown,
		sr44100,
		sr48000,
		sr88200,
		sr96000,
		sr176400,
		sr192000,
		sr32000,	// ordering for backwards compatibility
	};

	enum class AdcFilter : uint8_t
	{
		Normal,
		Slow = 0x01,		// slow roll-off
		Short = 0x02,		// short delay
		SlowShort = (Slow | Short),
	};

	enum class DacFilter : uint8_t
	{
		Normal,
		Slow = 0x01,		// slow roll-off
		Short = 0x02,		// short delay
		SlowShort = (Slow | Short),
		SuperSlow = 0x04,	// super slow roll-off
	};

	enum class EOffOn : uint8_t
	{
		Off,
		On
	};

	// Smart button modes.
	//
	// - Intelligent:
	//	When pressed shortly and released it toggles the state (Toggling).
	//	When pressed and held it becomes Momentary.
	//
	// - Momentary:
	//	The state is On as long as the button is being pressed.
	//
	// - Toggling:
	//	The state is toggled at every button press.
	//
	// - Cough:
	//	Inverse of Momentary - the normal state is On and it turns to Off as long as the button is pressed.
	//
	// - AlwaysOn:
	//	The state is always On, ignoring the button presses.
	//
	// - AlwaysOff:
	//	The state is always Off, ignoring the button presses.
	//
	enum class ButtonMode : uint8_t
	{
		Intelligent,
		Momentary,
		Toggling,
		Cough,
		AlwaysOn,
		AlwaysOff
	};

	// Volume and pan packed in two bytes
	//
	struct VolPan
	{
		Volume volume;
		Pan pan;
	};

	// Bit field that packs a Dante channel index and the direction in one byte.
	//
	struct DanteChannelId
	{
		uint8_t		index : 7;		// channel index 0 to 127
		bool		isOut : 1;		// false = IN, true = OUT
	};
	static_assert(sizeof(DanteChannelId) == 1);
	
	// String type.
	// Size is from 0 to 255, followed by that number of characters.
	// The characters must not include a '\0'.
	// Depending on context, encoding may be ASCII or UTF-8.
	//
	struct PascalString
	{
		uint8_t size;
		char characters[/*size*/];
	};
}
