#pragma once

/*
Copyright © Glensound Electronics Ltd and individual contributors. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met: Redistributions of source code must retain the
above copyright notice, this list of conditions and the following disclaimer. Redistributions
in binary form must reproduce the above copyright notice, this list of conditions and the
following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of Glensound nor the names of its contributors may be used to endorse or
promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Content based on github.com/github/opensource.guide used under the CC-BY-4.0 license.
*/

/*

 Firmware update protocol
 ========================

 NOTE: In all packets sent from the host to device for firmware update, the header must have
 the noMeter and exclusiveAccess bits set to 1. All reserved fields should be set to zero.

 Currently, this protocol only supports updating of the device's main firmware file (DFU).

 Conceptually, the host controller recognises four phases of the process: Idle, Erasing,
 Programming and Verifying.

 1) Host sends a SetUpdate message to the device. The payload is an UpdateHeader followed
 by a SectionDirectory structure. The update command is Initialise and the section directory
 lists all the firmware sections to be programmed.

 Header				- Opcode::SetUpdate, multipacket = 0, noMeter = 1, exclusiveAccess = 1
 UpdateHeader		- UpdateType::Dfu, UpdateCommand::Initialise
 SectionDirectory	- load address, length in words and CRC of all sections to be programmed
 [...the directory data...]

 After sending the message the host enters the Erasing phase.

 2) The device starts erasing its program memory and responds with one or more UpdaterReport
 messages, containing the update status, erasing progress (in percent), and:
 - the maximum number of 32-bit words that it can accept in one multi-packet transaction, and
 - an array reporting the number of words programmed for each firmware section.

 Header				- Opcode::UpdaterReport, multipacket = 0
 UpdateProgress		- progress information

 The UpdateStatus field of UpdateProgress indicates what the device is doing. If it is is
 Erasing the host can update the erase progress on the UI. The device returns Receiving
 status when it successfuly finished erasing. Any other status indicates a failure.

 3) After receiving an UpdaterReport message with Receiving status, the host should switch to
 Programming phase. The device is now waiting for program data.

 In Programming phase, if the host receives the Receiving status, it should scan the reported
 array of number of section words written, to find a section that is not fully written. If such
 section is found, the host sends the required section data remaining to program. The amount of
 data must not exceed the maximum number of words that the device can receive, as indicated in
 the maxBlockLengthWords field of the received report.

 Firmware section data block is sent to the device using one or more SetUpdate messages with
 the Program command in the UpdateHeader. The data payload is formatted as follows:

 UpdateHeader		- UpdateType::Dfu, UpdateCommand::Program
 UpdateData			- indicates the section number, address and length
 [...the section data...]

 - If this payload fits in a single packet then the packet has the following structure:
 
 Header				- Opcode::SetUpdate, multipacket = 0, noMeter = 1, exclusiveAccess = 1
 [payload]

 - Otherwise it is sent using a multi-packet transaction, with each packet as follows:

 Header				- Opcode::SetUpdate, multipacket = 1, noMeter = 1, exclusiveAccess = 1
 MultiPacketInfo	- contains info about the transaction and the current part number
 [partial payload]

 All parts of the transaction have maximum packet size, except possibly the last one.

 4) In Programming phase, if the host receives a Verifying status it enters the Verifying phase,
 in which it awaits a progress message from the device indicating success or failure.
 */

#include "GsTypes.h"


namespace Gs
{
	enum class UpdateType : uint8_t
	{
		Dnt,
		Dfu,	// currently, only this one is supported
		Dfu2
	};


	enum class UpdateCommand : uint8_t
	{
		Cancel,
		Initialise,
		Program
	};


	// Update status code returned by device in UpdateReport structure
	//
	enum class UpdateStatus : uint8_t
	{
		Idle = 0,
		Erasing,		// erasing in progress
		Receiving,		// waiting for section data to program
		reserved,
		Verifying,		// CRC verification in progress
		Success,		// finished programming with success
		Failure,		// unspecified failure
		BadFormat,		// incorrectly formatted data received from the host
		BadSequence,	// invalid sequence of commands received
		EraseFailed,	// Flash erasing operation failed
		ProgramFailed,	// Flash programming operation failed
		VerifyFailed,	// CRC of programmed Flash data incorrect
		Cancelled		// Cancel command was received
	};


	struct UpdateHeader
	{
		UpdateType		type;
		UpdateCommand	command;
		uint16_t		reserved;
	};
	static_assert(sizeof(UpdateHeader) == 4);


	// Firmware section info sent from host to device
	//
	struct FirmwareSection
	{
		uint32_t address;		// section starting address (32-bit-aligned)
		uint32_t lengthWords;	// section length in 32-bit words
		uint32_t crc;			// CRC32 of the section data
	};
	static_assert(sizeof(FirmwareSection) == 12);


	// List of firmware sections to program - sent from host to device
	//
	struct SectionDirectory
	{
		uint32_t		numSections;	// number of entries in the following array
		FirmwareSection	sections[];		// one struct per section
	};
	static_assert(sizeof(SectionDirectory) == 4);	// some compilers may fail this assertion due to zero-sized array

	constexpr uint32_t cMaxFirmwareSections = 8;
	constexpr uint cMaxSectionDirectorySize{ sizeof(SectionDirectory) + sizeof(FirmwareSection) * cMaxFirmwareSections };	// 100


	// Firmware section data chunk sent from host to device
	//
	struct UpdateData
	{
		uint32_t sectionNum;
		uint32_t address;
		uint32_t lengthWords;	// payload length in 32-bit words
		uint32_t payload[];
	};
	static_assert(sizeof(UpdateData) == 12);


	struct UpdateProgress
	{
		UpdateStatus	status;
		uint8_t			erasedPercent;			// erase progress (0 to 100)
		uint16_t		reserved;
		uint32_t		maxBlockLengthWords;	// maximum number of words per block transfer transaction
		uint32_t		numSections;			// number of entries in the array that follows
		// uint32_t		wordsWritten[];			//  one word per section; number of section words written so far
	};
	static_assert(sizeof(UpdateProgress) == 12);
}
