#pragma once

/*
Copyright © Glensound Electronics Ltd and individual contributors. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met: Redistributions of source code must retain the
above copyright notice, this list of conditions and the following disclaimer. Redistributions
in binary form must reproduce the above copyright notice, this list of conditions and the
following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of Glensound nor the names of its contributors may be used to endorse or
promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Content based on github.com/github/opensource.guide used under the CC-BY-4.0 license.
*/

#include "../Common/GsStatus.h"
#include "../Common/GsPresets.h"
#include "../Common/GsCommands.h"


namespace Gs::Dark1616M
{
	constexpr inline size_t cNumInputs = 16;
	constexpr inline size_t cNumOutputs = 16;
	constexpr inline size_t cNumDanteOutChannels = cNumInputs;	// device inputs go to Dante outputs
	constexpr inline size_t cNumDanteInChannels = cNumOutputs;	// device outputs come from Dante inputs
	constexpr inline size_t cNumMeterChannels{ cNumDanteInChannels + cNumDanteOutChannels };	// 32
	constexpr inline uint8_t cMaxGainIndexLine = 7;				// gain index range in Line mode (0 - 7)
	constexpr inline uint8_t cMaxGainIndexMic = 17;				// gain index range in Mic mode (0 - 17)
	constexpr inline uint8_t cMaxGainIndexPhantom = 17;			// gain index range in Phantom mode (0 - 17)

	// Device-specific report types
	//
	enum class ReportId : uint8_t
	{
		_first = Gs::cNextReportId,
		IoBoard = _first,
		MicInBoard,
		_end
	};
	constexpr inline uint cNumReportTypes = (uint)ReportId::_end - (uint)ReportId::_first;						// 2
	constexpr inline uint cMaxReportTypes = (cNumReportTypes + sizeof(uint32_t) - 1) & ~(sizeof(uint32_t) - 1);	// 4 (with padding)

	// Device-specific status
	//
	struct Status : Gs::Status
	{
		uint8_t 	reportGenerations[cMaxReportTypes];		// [0]: IoBoard rep. gen, [1]: MicInBoard rep. gen.

		// Meter data:
		// 0-15		DOUT0-15
		// 16-31	DIN0-15

		uint8_t		meters[cNumMeterChannels];	// 32 bytes
		uint8_t		aes3LockStates;				// 1 bit per stereo pair, 0 = unlocked, 1 = locked
		uint8_t		selectedInputs;				// 1 bit per stereo pair, 0 = analogue, 1 = digital
		uint16_t	reserved;					// set to 0
	};
	static_assert(sizeof(Status) == 48);
	static_assert(sizeof(Status) + sizeof(Header) == 64);

	// Device-specific commands
	//
	// The following standard commands are also supported (GsCommands.h):
	// - AdcSettings
	// - DacSettings
	// - RestoreDefaults
	// - ChannelName
	//
	enum class Command : uint8_t
	{
		// 4-byte commands
		//
		GainIndex = Gs::cNext4byteCommand,	// message is GainIndexMsg
		InputMode,							// message is InputModeMsg
		InputTypeChoice,					// message is InputChoiceMsg
		ChannelPairing						// message is ChannelPairingMsg
	};

	struct GainIndexMsg
	{
		Command		command;	// Command::GainIndex
		uint8_t		index;		// channel index (0 to 15)
		GainIndex	gainIndex;	// Line: 0-7, Mic: 0-17, Mic+Phantom: 0-17
		uint8_t		reserved;
	};
	static_assert(sizeof(GainIndexMsg) == 4);

	struct InputModeMsg
	{
		Command				command;	// Command::InputMode
		uint8_t				index;		// channel index (0 to 15)
		AnalogueInputMode	mode;		// 0 = reserved, 1 = line, 2 = mic, 3 = phantom
		uint8_t				reserved;
	};
	static_assert(sizeof(InputModeMsg) == 4);

	struct InputChoiceMsg
	{
		Command		command;	// Command::InputChoice
		uint8_t		index;		// channel pair index (0 to 7)
		InputChoice	choice;		// 0 = auto, 1 = analogue, 2 = digital
		uint8_t		reserved;
	};
	static_assert(sizeof(InputChoiceMsg) == 4);

	struct ChannelPairingMsg
	{
		Command		command;	// Command::ChannelPairing
		uint8_t		index;		// channel pair index (0 to 7)
		Pairing		pairing;	// 0 = mono, 1 = stereo
		uint8_t		reserved;
	};
	static_assert(sizeof(ChannelPairingMsg) == 4);
	
	struct IoBoardReportData
	{
		uint16_t inputChoices;	// two bits per input pair: 00 = auto, 01 = analogue, 10 = digital
		uint16_t reserved;
	};
	static_assert(sizeof(IoBoardReportData) == 4);

	struct alignas(uint32_t) MicInBoardReportData
	{
		using ChannelGains = std::array<uint8_t, 4>;		// one gain per input mode 1-3 (input mode 0 is reserved)
		using Gains = std::array<ChannelGains, cNumInputs>;	// one set of gains per mic input

		Gains		gains;
		uint32_t	inputModes;		// two bits per mic input: 00 = reserved, 01 = Line, 10 = Mic, 11 = Mic+phantom
		uint8_t		pairings;		// one bit per channel pair: 0 = dual mono, 1 = stereo
		uint8_t		reserved1;		// set to zero
		uint16_t	reserved2;		// set to zero
	};
	static_assert(sizeof(MicInBoardReportData) == 72);

	// Settings types that can be reset individually with the RestoreDefaults command
	//
	enum class SettingsType : uint8_t
	{
		Channels = Gc::cNextPresetType
	};

	// Flags to be ORed to define a selection of settings to be restored
	//
	enum class SettingsPart : uint32_t
	{
		Channels = 1 << (int)SettingsType::Channels
	};

	// Preset types that can be applied individually
	//
	enum class PresetType : uint8_t
	{
		InputChoices = Gc::cNextPresetType,
		InputSettings
	};

	// Flags to be ORed to define a selection of preset parts to be applied
	//
	enum class PresetPart : uint32_t
	{
		InputChoices = 1 << (int)PresetType::InputChoices,
		InputSettings = 1 << (int)PresetType::InputSettings
	};
}
