#pragma once

/*
Copyright © Glensound Electronics Ltd and individual contributors. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met: Redistributions of source code must retain the
above copyright notice, this list of conditions and the following disclaimer. Redistributions
in binary form must reproduce the above copyright notice, this list of conditions and the
following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of Glensound nor the names of its contributors may be used to endorse or
promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Content based on github.com/github/opensource.guide used under the CC-BY-4.0 license.
*/

#include "../Common/GsStatus.h"
#include "../Common/GsPresets.h"
#include "../Common/GsCommands.h"


namespace Gs::Divine
{
	// Device-specific report types
	//
	enum class ReportId : uint8_t
	{
		_first = Gs::cNextReportId,
		All = _first,				// 4
		_end
	};
	constexpr uint cNumReportTypes = (uint)ReportId::_end - (uint)ReportId::_first;							// 1
	constexpr uint cMaxReportTypes = (cNumReportTypes + sizeof(uint32_t) - 1) & ~(sizeof(uint32_t) - 1);	// 4 (with padding)


	// Device-specific status
	//
	struct Status : Gs::Status
	{
		uint8_t reportGenerations[cMaxReportTypes];

		// Audio meter peak values (Dante format):
		//
		// 0: IN 1
		// 1: IN 2
		// 2: IN 3
		// 3: IN 4
		// 4: IN 1+2
		// 5: IN 3+4
		// 6: IN 1+2+3+4
		// 7: speaker output
		//
		uint8_t meters[8];

		uint8_t potValue;		// pot position 0 - 127
		uint8_t volume;			// master volume index 0 - 127
		int8_t temperature;		// internal temperature in Celsius, T = 0.5 * temperature + 44
		uint8_t reserved;
	};
	static_assert(sizeof(Status) == 24);
	static_assert(sizeof(Status) + sizeof(Header) == 40);


	// The following standard commands are supported (GsCommands.h):
	// - DacSettings
	// - RestoreDefaults
	// - ChannelName

	// Device-specific commands
	//
	enum class Command : uint8_t
	{
		// 4-byte commands
		//
		MixSelection = Gs::cNext4byteCommand,	// 5 MixSelectionMsg
		MixEnable,			// 6 MixEnableMsg
		ChannelDelay,		// 7 ControlMsg, val1 = channel (0 - 3), val2 = delayMs (0 - 40)
		ChannelTrim,		// 8 ChannelTrimMsg
		ChannelFilter,		// 9 ChannelFilterMsg
		ChannelPriority,	// 10 ChannelPriorityMsg
		SoloDim,			// 11 SoloDimMsg
		MinSolo,			// 12 MinSoloMsg
		Lineup,				// 13 LineupMsg
		Volume,				// 14 ControlMsg, val1 = gain index (0 - 127)
		MinVolume,			// 15 MinVolumeMsg
		VolumeLimit,		// 16 VolumeLimitMsg
		VolumeBoost,		// 17 VolumeBoostMsg
		EqType,				// 18 EqTypeMsg
		LfCut,				// 19 OffOnMsg
		PotEnable,			// 20 OffOnMsg
		MixSelectionEnable,	// 21 OffOnMsg
		RearButtonsEnable,	// 22 OffOnMsg
		RearPasswordEnable,	// 23 OffOnMsg
		SetPassword,		// 24 ControlMsg, val1 = password (0 - 255)
		DarkMode,			// 25 OffOnMsg
		PowerSave,			// 26 PowerSaveMsg
		MeterMode,			// 27 MeterModeMsg
		DisplayBrightness,	// 28 ControlMsg, val1 = brightness (0 - 100)
		SpeakerGroup,		// 29 ControlMsg, val1 = group number (0 - 10)
		PeerMixSelection,	// 30 MixSelectionMsg
		PeerVolume,			// 31 ControlMsg, val1 = gain index (0 - 127)
		PhaseInvert			// 32 OffOnMsg
	};


	// Command-specific data structures and definitions

	enum class ChannelNumber : uint8_t
	{
		None,
		Ch1,
		Ch2,
		Ch3,
		Ch4
	};

	enum class MixType : uint8_t
	{
		None,
		Ch1,
		Ch2,
		Ch3,
		Ch4,
		Ch1_2,
		Ch3_4,
		Ch1_4,
		_NumOptions
	};

	// At 48 kHz all mix types are allowed
	//
	constexpr bool cValidMixes48[(int)MixType::_NumOptions]
	{
		false,		// None
		true,		// Ch1
		true,		// Ch2
		true,		// Ch3
		true,		// Ch4
		true,		// Ch1_2
		true,		// Ch3_4
		true		// Ch1_4
	};

	// At 96 kHz only two input channels are supported
	//
	constexpr bool  cValidMixes96[(int)MixType::_NumOptions]
	{
		false,		// None
		true,		// Ch1
		true,		// Ch2
		false,		// Ch3
		false,		// Ch4
		true,		// Ch1_2
		false,		// Ch3_4
		false		// Ch1_4
	};

	// Bit map indicating which mixes are enabled
	//
	enum class MixEnable : uint8_t
	{
		None = 1 << (int)MixType::None,
		Ch1 = 1 << (int)MixType::Ch1,
		Ch2 = 1 << (int)MixType::Ch2,
		Ch3 = 1 << (int)MixType::Ch3,
		Ch4 = 1 << (int)MixType::Ch4,
		Ch1_2 = 1 << (int)MixType::Ch1_2,
		Ch3_4 = 1 << (int)MixType::Ch3_4,
		Ch1_4 = 1 << (int)MixType::Ch1_4,
	};

	enum class Lineup : uint8_t
	{
		Minus18,	// normal (-18 dBFS), extra gain of 0 dB
		Minus20,	// -20 dBFs, extra gain of 2 dB
		Minus24		// -24 dBFs, extra gain of 6 dB
	};

	enum class VolumeBoost : uint8_t
	{
		Plus0,
		Plus6,
		Plus12,
		Plus18,
		Plus24
	};

	enum class VolumeLimit : uint8_t
	{
		Minus0,
		Minus3,
		Minus6,
		Minus9,
		Minus12
	};

	enum class MinVolume : uint8_t
	{
		Mute,
		Minus52,
		Minus46,
		Minus40,
		Minus34,
		Minus28
	};

	enum class SoloDim : uint8_t
	{
		Minus6,
		Minus12,
		Minus18,
		Minus24,
		Minus30,
		Mute
	};

	enum class MinSolo : uint8_t
	{
		Minus24,
		Minus30,
		Minus36,
		Minus42,
		Minus48,
		Minus54,
		Minus58,
		Mute
	};

	enum class EqType : uint8_t
	{
		Off,
		Natural,
		Normal,
		Pop
	};

	enum class MeterMode : uint8_t
	{
		Off,
		Bar,
		Colours
	};

	enum class ChannelPriority : uint8_t
	{
		P1,
		P2,
		P3
	};

	enum class PowerSave : uint8_t
	{
		Off,
		Seconds10,
		Seconds30,
		Seconds60,
		Minutes5,
		Minutes15,
		Minutes30,
		Minutes60
	};

	struct alignas(uint32_t) MixSelectionMsg	// 4 bytes
	{
		Command command;				// Command::MixSelection
		MixType mix;
		uint8_t reserved[2];
	};
	static_assert(sizeof(MixSelectionMsg) == 4);

	struct alignas(uint32_t) MixEnableMsg	// 4 bytes
	{
		Command command;				// Command::MixEnable
		MixType mix;
		EOffOn state;
		uint8_t reserved;
	};
	static_assert(sizeof(MixEnableMsg) == 4);

	struct alignas(uint32_t) LineupMsg		// 4 bytes
	{
		Command command;				// Command::Lineup
		Lineup lineup;
		uint8_t reserved[2];
	};
	static_assert(sizeof(LineupMsg) == 4);

	struct alignas(uint32_t) VolumeLimitMsg	// 4 bytes
	{
		Command command;				// Command::VolumeLimit
		VolumeLimit limit;
		uint8_t reserved[2];
	};
	static_assert(sizeof(VolumeLimitMsg) == 4);

	struct alignas(uint32_t) MinVolumeMsg	// 4 bytes
	{
		Command command;				// Command::MinVolume
		MinVolume limit;
		uint8_t reserved[2];
	};
	static_assert(sizeof(MinVolumeMsg) == 4);

	struct alignas(uint32_t) VolumeBoostMsg	// 4 bytes
	{
		Command command;				// Command::VolumeBoost
		VolumeBoost boost;
		uint8_t reserved[2];
	};
	static_assert(sizeof(VolumeBoostMsg) == 4);

	struct alignas(uint32_t) SoloDimMsg	// 4 bytes
	{
		Command command;				// Command::SoloDim
		SoloDim dim;
		uint8_t reserved[2];
	};
	static_assert(sizeof(SoloDimMsg) == 4);

	struct alignas(uint32_t) MinSoloMsg	// 4 bytes
	{
		Command command;				// Command::MinSolo
		MinSolo limit;
		uint8_t reserved[2];
	};
	static_assert(sizeof(MinSoloMsg) == 4);

	struct alignas(uint32_t) ChannelTrimMsg	// 4 bytes
	{
		Command command;				// Command::ChannelTrim
		uint8_t channel;				// 0 to 3
		int8_t trim;					// -64 to 63
		uint8_t reserved;
	};
	static_assert(sizeof(ChannelTrimMsg) == 4);

	struct alignas(uint32_t) ChannelFilterMsg	// 4 bytes
	{
		Command command;				// Command::ChannelFilter
		uint8_t channel;				// 0 to 3
		EOffOn filter;					// 0=disabled, 1=speech
		uint8_t reserved;
	};
	static_assert(sizeof(ChannelFilterMsg) == 4);

	struct alignas(uint32_t) ChannelPriorityMsg	// 4 bytes
	{
		Command command;				// Command::ChannelPriority
		ChannelPriority priority;		// 0=P1, 1=P3, 2=P3
		ChannelNumber channel;			// which channel has the above priority
		uint8_t reserved;
	};
	static_assert(sizeof(ChannelPriorityMsg) == 4);

	struct alignas(uint32_t) EqTypeMsg		// 4 bytes
	{
		Command command;				// Command::EqType
		MixType mix;
		EqType eq;
		uint8_t reserved;
	};
	static_assert(sizeof(EqTypeMsg) == 4);

	struct alignas(uint32_t) PowerSaveMsg	// 4 bytes
	{
		Command command;				// Command::PowerSave
		PowerSave time;
		uint8_t reserved[2];
	};
	static_assert(sizeof(PowerSaveMsg) == 4);

	struct alignas(uint32_t) MeterModeMsg	// 4 bytes
	{
		Command command;				// Command::MeterMode
		MeterMode mode;
		uint8_t reserved[2];
	};
	static_assert(sizeof(MeterModeMsg) == 4);

	struct alignas(uint32_t) OffOnMsg
	{
							// Command::PhaseInvert
							// Command::DarkMode
							// Command::RearPasswordEnable
							// Command::RearButtonsEnable
							// Command::MixSelectionEnable
							// Command::PotEnable
		Command command;	// Command::LfCut
		EOffOn state;
		uint8_t param1;
		uint8_t param2;
	};
	static_assert(sizeof(OffOnMsg) == 4);

	struct alignas(uint32_t) ControlMsg	// 4 bytes
	{
							// Command::PeerVolume			val1 = gain index (0 - 127)
							// Command::SpeakerGroup		val1 = group number (0 - 10)
							// Command::DisplayBrightness	val1 = brightness (0 - 100)
							// Command::SetPassword			val1 = password (0 - 255)
							// Command::Volume				val1 = gain index (0 - 127)
		Command command;	// Command::ChannelDelay		val1 = channel (0 - 3), val2 = delayMs (0 - 40)
		uint8_t val1;
		uint8_t val2;
		uint8_t val3;
	};
	static_assert(sizeof(ControlMsg) == 4);


	// Report data
	//
	struct alignas(uint32_t) ReportData
	{
		struct
		{
			uint32_t lfCut : 1;			// 0=disable, 1=enable
			uint32_t pot : 1;			// 0=disable, 1=enable
			uint32_t mixSelection : 1;	// 0=disable, 1=enable
			uint32_t darkMode : 1;		// 0=disable, 1=enable
			uint32_t rearButtons : 1;	// 0=disable, 1=enable
			uint32_t rearPassword : 1;	// 0=disable, 1=enable
			uint32_t phaseInvert : 1;	// 0=disable, 1=enable
			uint32_t reserved : 25;
		} flags;
		struct
		{
			uint32_t reserved : 4;
			uint32_t eq1 : 4;			// MixType::Ch1
			uint32_t eq2 : 4;			// MixType::Ch2
			uint32_t eq3 : 4;			// MixType::Ch3
			uint32_t eq4 : 4;			// MixType::Ch4
			uint32_t eq5 : 4;			// MixType::Ch1_2
			uint32_t eq6 : 4;			// MixType::Ch3_4
			uint32_t eq7 : 4;			// MixType::Ch1_4
		} eqTypes;
		struct
		{
			uint16_t priority1 : 3;		// channel number for priority 1
			uint16_t priority2 : 3;		// channel number for priority 2
			uint16_t priority3 : 3;		// channel number for priority 3
		} priorities;
		SoloDim		dim;
		MinSolo		minSolo;
		int8_t		channelTrims[4];	// gain trim values per channel (-64 to 63)
		uint8_t		delays[4];			// delay per channel, in ms (0 to 40)
		uint8_t		password;
		MeterMode	meterMode;
		VolumeBoost	volumeBoost;
		VolumeLimit	volumeLimit;
		MinVolume	minVolume;
		Lineup		lineup;
		PowerSave	powerSave;
		MixType		mixType;			// selected mix type
		MixEnable	mixEnables;			// enabled mix types bitmap
		struct
		{
			uint8_t channel1 : 1;
			uint8_t channel2 : 1;
			uint8_t channel3 : 1;
			uint8_t channel4 : 1;
		} speechFilterEnables;					// a bit of 1 indicates speech filter enabled in this channel
		uint8_t		displayBrightnessPercent;	// 0 - 100
		uint8_t		speakerGroup;				// 0 - 15
		uint8_t		reserved[4];
	};
	static_assert(sizeof(ReportData) == 36);


	// Device-specific report transaction (single packet)
	//
	struct ReportPacket
	{
		Header header;				// .size = sizeof(ReportPacket)
									// .multipacket = 0
									// .opcode = Opcode::Report

		BlockHeader blockHeader;	// .type = ReportId::All
									// .param = report generation number
									// .payloadWords = sizeof(ReportData) / sizeof(uint32_t)

		ReportData data;
	};
	static_assert(sizeof(ReportPacket) == 56);
}
