#pragma once

/*
Copyright  Glensound Electronics Ltd and individual contributors. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met: Redistributions of source code must retain the
above copyright notice, this list of conditions and the following disclaimer. Redistributions
in binary form must reproduce the above copyright notice, this list of conditions and the
following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of Glensound nor the names of its contributors may be used to endorse or
promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS AS IS AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Content based on github.com/github/opensource.guide used under the CC-BY-4.0 license.
*/

#include "../Common/GsStatus.h"
#include "../Common/GsPresets.h"
#include "../Common/GsCommands.h"


namespace Gs::Spark
{
	// Device-specific report types
	//
	enum class ReportId : uint8_t
	{
		_first = Gs::cNextReportId,
		Ui = _first,
		Mic,
		_end
	};
	constexpr inline uint cNumReportTypes{ (uint)ReportId::_end - (uint)ReportId::_first };							// 2
	constexpr inline uint cMaxReportTypes{ (cNumReportTypes + sizeof(uint32_t) - 1) & ~(sizeof(uint32_t) - 1) };	// 4 (with padding)

	// Device-specific status
	//
	struct Status : Gs::Status
	{
		uint8_t reportGenerations[cMaxReportTypes];

		// Audio meter peak values (Dante format):
		//
		// 0..7:	Dante outputs 1..8
		// 8..15:	Dante inputs 1..8
		// 16:		Mic input
		// 17..19:	reserved
		//
		uint8_t meters[20];

		uint8_t leds;			// LED states, 5 bits used, see LedId
		uint8_t reserved1;
		uint8_t reserved2;
		int8_t temperature;		// internal temperature in Celsius, T = 0.5 * temperature + 44
	};
	static_assert(sizeof(Status) == 36);
	static_assert(sizeof(Status) + sizeof(Header) == 52);

	// The following standard commands are supported (GsCommands.h):
	// - AdcSettings
	// - DacSettings
	// - RestoreDefaults
	// - ChannelName

	// Device-specific commands
	//
	enum class Command : uint8_t
	{
		// 4-byte commands
		//
		MicGain = Gs::cNext4byteCommand,	// 5 Value8, 0-63
		Phantom,		// 6 OffOnMsg
		Compressor,		// 7 OffOnMsg
		PotVolume,		// 8 VolumeMsg
		PotPan,			// 9 PanMsg
		Button,			// 10 OffOnMsg
		ButtonMode,		// 11 ButtonModeMsg
		PgmMuteEnable,	// 12 PgmMuteEnableMsg
		PotMode,		// 13 PotModeMsg
		MicGainLock,	// 14 OffOnMsg
		PhantomLock		// 15 OffOnMsg
	};

	// Command-specific data structures and definitions

	enum class Channel : uint8_t
	{
		Cue1,
		Cue2,
		Cue3,
		Cue4,
		Hp,
		Sidetone,
		Aux,
		Background,
		_NumChannels
	};

	enum class PotMode : uint8_t
	{
		Unlimited,
		Limited,
	};

	enum class ButtonId : uint8_t
	{
		Pgm,
		Tb1,
		Tb2,
		Tb3,
		Alt,
		Pan
	};

	enum class PotId : uint8_t
	{
		Pot1,
		Pot2,
		Pot3,
		Pot4,
		MicGain
	};

	// Bit indices in the Status::leds field
	//
	enum class LedId : uint8_t
	{
		Pgm,
		Tb1,
		Tb2,
		Tb3,
		Power
	};

	struct OffOnMsg
	{
							// Command::Button
							// Command::MicGainLock
							// Command::PhantomLock
							// Command::Compressor
		Command	command;	// Command::Phantom
		EOffOn	state;
		uint8_t	index;		// only used for Command::Button
		uint8_t	reserved2;
	};
	static_assert(sizeof(OffOnMsg) == 4);

	struct VolumeMsg
	{
		Command		command;		// Command::Volume
		Volume		volume;			// 0 to 127 (0.5 dB steps, 104 is 0 dB)
		Channel		channel;		// 0 to 7
		uint8_t		reserved;
	};
	static_assert(sizeof(VolumeMsg) == 4);

	struct PanMsg
	{
		Command		command;		// Command::Pan
		Pan			pan;			// -7 to 7
		Channel		channel;		// 0 to 7
		uint8_t		reserved;
	};
	static_assert(sizeof(PanMsg) == 4);

	struct ButtonModeMsg
	{
		Command		command;		// Command::ButtonMode
		ButtonId	buttonId;		// 0 to 3
		ButtonMode	mode;
		uint8_t		reserved;
	};
	static_assert(sizeof(ButtonModeMsg) == 4);

	struct PotModeMsg
	{
		Command		command;		// Command::PotMode
		Channel		channel;		// 0 to 7
		PotMode		mode;
		uint8_t		reserved;
	};
	static_assert(sizeof(PotModeMsg) == 4);

	struct PgmMuteEnableMsg
	{
		Command		command;		// Command::PgmMuteEnable
		ButtonId	buttonId;		// 1 to 3 (only TB1..3 buttons may mute PGM)
		uint8_t		enable;			// 0=disable, 1=enable
		uint8_t		reserved;
	};
	static_assert(sizeof(PgmMuteEnableMsg) == 4);

	// Mic input report
	//
	struct MicInReport
	{
		uint8_t gainIndex;		// 0 to 63 (maps to -8..55 dB)
		union
		{
			struct
			{
				bool compressor : 1;	// 0=disable, 1=enable
				bool phantom : 1;		// 0=disable, 1=enable
				bool gainLocked : 1;	// 0=disable, 1=enable
				bool phantomLocked : 1;	// 0=disable, 1=enable
				bool reserved : 4;
			};
			uint8_t flags;
		} flags;
		uint16_t reserved;
	};
	static_assert(sizeof(MicInReport) == 4);

	// UI settings report
	//
	struct UiReport
	{
		struct
		{
			ButtonMode pgmMode : 4;		// mode of PGM button
			ButtonMode tb1Mode : 4;		// mode of TB1 button
			ButtonMode tb2Mode : 4;		// mode of TB2 button
			ButtonMode tb3Mode : 4;		// mode of TB3 button
		};
		struct
		{
			PotMode cue1Mode : 2;		// mode of Cue1 pot
			PotMode cue2Mode : 2;		// mode of Cue2 pot
			PotMode cue3Mode : 2;		// mode of Cue3 pot
			PotMode cue4Mode : 2;		// mode of Cue4 pot
			PotMode hpMode : 2;			// mode of HP pot
			PotMode sidetoneMode : 2;	// mode of Sidetone pot
			PotMode auxMode : 2;		// mode of Aux In pot
			PotMode bkgMode : 2;		// mode of Background In pot
		};
		VolPan volPans[(int)Channel::_NumChannels];	// volume and pan of each channel
		struct
		{
			bool reserved1 : 1;
			bool tb1MutesPgm : 1;		// 0=disable, 1=enable
			bool tb2MutesPgm : 1;		// 0=disable, 1=enable
			bool tb3MutesPgm : 1;		// 0=disable, 1=enable
		};
		uint8_t reserved2[3];
	};
	static_assert(sizeof(UiReport) == 24);	// NOTE: some compilers may not agree!
	
	// Settings types that can be reset individually with the RestoreDefaults command
	//
	enum class SettingsType : uint8_t
	{
		Ui = Gs::Protocol::cNextPresetType,
		Mic
	};

	// Flags to be ORed to define a selection of settings to be restored
	//
	enum class SettingsPart : uint32_t
	{
		Ui =  1 << (int)SettingsType::Ui,
		Mic =  1 << (int)SettingsType::Mic,
	};

	// Preset types that can be applied individually
	//
	enum class PresetType : uint8_t
	{
		Ui = Gs::Protocol::cNextPresetType,
		Mic
	};

	// Flags to be ORed to define a selection of preset parts to be applied
	//
	enum class PresetPart : uint32_t
	{
		Ui = 1 << (int)PresetType::Ui,
		Mic = 1 << (int)PresetType::Mic,
	};
}
