local helpers = require("Helpers")

local m =
{
	-- Status
	identify = ProtoField.bool("gs.device.identify", "Identify", 32, nil, 0x0001),
	clkLocked = ProtoField.bool("gs.device.clkLocked", "Clock locked", 32, nil, 0x0002),
	presetRepGen = ProtoField.uint8("gs.device.presetRepGen", "Preset rep. gen.", base.DEC),
	deviceRepGen = ProtoField.uint8("gs.device.deviceRepGen", "Device rep. gen.", base.DEC),
	inChNamesRepGen = ProtoField.uint8("gs.device.inChNamesRepGen", "In ch. names rep. gen.", base.DEC),
	outChNamesRepGen = ProtoField.uint8("gs.device.outChNamesRepGen", "Out ch. names rep. gen.", base.DEC),

	-- Report/Command
	clkSrc = ProtoField.uint8("gs.device.clkSrc", "Clock source", base.DEC),
	adcFilter = ProtoField.uint8("gs.device.adcFilter", "ADC filter", base.DEC),
	dacFilter = ProtoField.uint8("gs.device.dacFilter", "DAC filter", base.DEC),

	-- Report
	sampleRate = ProtoField.uint32("gs.device.sampleRate", "Sample rate", base.DEC),
	srateChoice = ProtoField.uint8("gs.device.srateChoice", "Sample rate choice", base.DEC),
	clkSrcChoice = ProtoField.uint8("gs.device.clkSrcChoice", "Clock source choice", base.DEC),
	link1 = ProtoField.bool("gs.device.link1", "Link 1", 32, nil, 0x00000001),
	link2 = ProtoField.bool("gs.device.link2", "Link 2", 32, nil, 0x00000002),
	psuState1 = ProtoField.bool("gs.device.psuState1", "PSU 1", 32, nil, 0x00000004),
	psuState2 = ProtoField.bool("gs.device.psuState2", "PSU 2", 32, nil, 0x00000008),
	poeState1 = ProtoField.bool("gs.device.poeState1", "POE 1", 32, nil, 0x00000010),
	poeState2 = ProtoField.bool("gs.device.poeState2", "POE 2", 32, nil, 0x00000020),

	-- Command
	srate = ProtoField.uint8("gs.device.srate", "Srate", base.DEC),
	settingsParts = ProtoField.uint32("gs.device.settingsParts", "Settings parts", base.DEC),
}


function m.GetFields()
	local fields =
	{
		-- Status
		identify = m.identify,
		clkLocked = m.clkLocked,
		presetRepGen = m.presetRepGen,
		deviceRepGen = m.deviceRepGen,
		inChNamesRepGen = m.inChNamesRepGen,
		outChNamesRepGen = m.outChNamesRepGen,

		-- Report/Command
		clkSrc = m.clkSrc,
		adcFilter = m.adcFilter,
		dacFilter = m.dacFilter,

		-- Report
		sampleRate = m.sampleRate,
		srateChoice = m.srateChoice,
		clkSrcChoice = m.clkSrcChoice,
		link1 = m.link1,
		link2 = m.link2,
		psuState1 = m.psuState1,
		psuState2 = m.psuState2,
		poeState1 = m.poeState1,
		poeState2 = m.poeState2,

		-- Command
		srate = m.srate,
		settingsParts = m.settingsParts,
	}
	
	return fields
end


local fieldSrateChoice
local fieldClkSrcChoice
local fieldClkSrc
local fieldAdcFilter
local fieldDacFilter
local fieldSrate


function m.OnFieldsRegistered()
	-- Report
	fieldSrateChoice = Field.new("gs.device.srateChoice")
	fieldClkSrcChoice = Field.new("gs.device.clkSrcChoice")

	-- Report/Command
	fieldClkSrc = Field.new("gs.device.clkSrc")
	fieldAdcFilter = Field.new("gs.device.adcFilter")
	fieldDacFilter = Field.new("gs.device.dacFilter")
	
	-- Command
	fieldSrate = Field.new("gs.device.srate")
end


local function GetSrateName(srate)
	if srate == 1 then return "44100"
	elseif srate == 2 then return "48000"
	elseif srate == 3 then return "88200"
	elseif srate == 4 then return "96000"
	elseif srate == 5 then return "176400"
	elseif srate == 6 then return "192000"
	elseif srate == 7 then return "32000"
	end
	
	return "Unknown"
end


local function GetClkSrcName(source)
	if source == 0 then return "Dante"
	elseif source == 1 then return "ExtWck"
	elseif source == 2 then return "Input"
	elseif source == 2 then return "Crystal"
	end

	return "Invalid"
end


local function GetAdcFilterName(filter)
	if filter == 0 then return "Normal"
	elseif filter == 1 then return "Slow"
	elseif filter == 2 then return "Short"
	elseif filter == 3 then return "SlowShort"
	end

	return "Invalid"
end


local function GetDacFilterName(filter)
	if filter == 0 then return "Normal"
	elseif filter == 1 then return "Slow"
	elseif filter == 2 then return "Short"
	elseif filter == 3 then return "SlowShort"
	elseif filter == 4 then return "SuperSlow"
	end

	return "Invalid"
end


local function GetCmdName(cmd)
	if cmd == 1 then return "Sample rate"
	elseif cmd == 2 then return "Clock source"
	elseif cmd == 3 then return "ADC settings"
	elseif cmd == 4 then return "DAC settings"
	elseif cmd == 128 then return "Restore settings"
	elseif cmd == 192 then return "Channel name"
	end
	
	return "Unknown"
end


function m.ParseStatus(proto, buffer, tree)
	tree = tree:add(proto, buffer(0,8), "GS Status")
	local bytes = buffer(0,4)
	local subtree = tree:add_le(g_f.flags32, bytes)
	subtree:add_le(m.identify, bytes)
	subtree:add_le(m.clkLocked, bytes)

	tree:add(m.presetRepGen, buffer(4,1))
	tree:add(m.deviceRepGen, buffer(5,1))
	tree:add(m.inChNamesRepGen, buffer(6,1))
	tree:add(m.outChNamesRepGen, buffer(7,1))
end


-- Returns the number of consumed bytes
function m.ParseReport(proto, buffer, reportSize, tree)
	local size = buffer:len()
	
	if size < 16 or reportSize ~= 16 then
		helpers.ErrorSize(tree)
		return size
	end
	
	tree = tree:add(proto, buffer(0,16), "Device Report")
	
	tree:add_le(m.sampleRate, buffer(0,4))
	local t = tree:add(m.srateChoice, buffer(4,1))
	t:append_text(" ("..GetSrateName(fieldSrateChoice()())..")")
	t = tree:add(m.clkSrc, buffer(5,1))
	t:append_text(" ("..GetClkSrcName(fieldClkSrc()())..")")
	t = tree:add(m.clkSrcChoice, buffer(6,1))
	t:append_text(" ("..GetClkSrcName(fieldClkSrcChoice()())..")")
	t = tree:add(m.adcFilter, buffer(7,1))
	t:append_text(" ("..GetAdcFilterName(fieldAdcFilter()())..")")
	t = tree:add(m.dacFilter, buffer(8,1))
	t:append_text(" ("..GetDacFilterName(fieldDacFilter()())..")")
	local bytes = buffer(12,4)
	t = tree:add_le(g_f.flags32, bytes)
	t:add_le(m.link1, bytes)
	t:add_le(m.link2, bytes)
	t:add_le(m.psuState1, bytes)
	t:add_le(m.psuState2, bytes)
	t:add_le(m.poeState1, bytes)
	t:add_le(m.poeState2, bytes)
	
	return 16
end


function m.ParseCommand(buffer, tree)	
	local t
	local cmd = buffer(0, 1):uint()
	local name = GetCmdName(cmd)
	tree:append_text(" ("..name..")")
	
	if cmd == 1 then
		t = tree:add(m.srate, buffer(1, 1))
		t:append_text(" ("..GetSrateName(fieldSrate()())..")")
	elseif cmd == 2 then
		t = tree:add(m.clkSrc, buffer(1, 1))
		t:append_text(" ("..GetClkSrcName(fieldClkSrc()())..")")
	elseif cmd == 3 then
		t = tree:add(m.adcFilter, buffer(1, 1))
		t:append_text(" ("..GetAdcFilterName(fieldAdcFilter()())..")")
	elseif cmd == 4 then
		t = tree:add(m.dacFilter, buffer(1, 1))
		t:append_text(" ("..GetDacFilterName(fieldDacFilter()())..")")
	elseif cmd == 128 then
		tree:add(m.settingsParts, buffer(4, 4))
	elseif cmd == 192 then
		-- channel name
	end
end

return m
