local m =
{
	headerMagic = ProtoField.bytes("gs.headerMagic", "Magic"),
	packetSize = ProtoField.uint16("gs.packetSize", "Packet size", base.DEC, nil, 0x7fff, "including this header"),
	isMultipacket = ProtoField.bool("gs.isMultipacket", "Multipacket", 16, nil, 0x8000),
	isExclusive = ProtoField.bool("gs.isExclusive", "Exclusive", 8, nil, 0x01),
	isNoMeter = ProtoField.bool("gs.isNoMeter", "No meter", 8, nil, 0x02),
	isPasswordValid = ProtoField.bool("gs.isPasswordValid", "Password valid", 8, nil, 0x04),
	isAccessGranted = ProtoField.bool("gs.isAccessGranted", "Access granted", 8, nil, 0x08),
	controllerId = ProtoField.uint32("gs.controllerId", "Controller ID", base.HEX),
	sequence = ProtoField.uint8("gs.sequence", "Sequence", base.DEC),
	devId = ProtoField.bytes("gs.devId", "Device ID")
}


function m.GetFields()
	local fields =
	{
		headerMagic = m.headerMagic,
		packetSize = m.packetSize,
		isMultipacket = m.isMultipacket,
		isExclusive = m.isExclusive,
		isNoMeter = m.isNoMeter,
		isPasswordValid = m.isPasswordValid,
		isAccessGranted = m.isAccessGranted,
		controllerId = m.controllerId,
		sequence = m.sequence,
		devId = m.devId
	}
	
	return fields
end


local function GetName(opcode)
	if opcode == 0 then return "Reserved"
	elseif opcode == 1 then return "Status"
	elseif opcode == 2 then return "GetStatus"
	elseif opcode == 3 then return "SetControl"
	elseif opcode == 4 then return "Info"
	elseif opcode == 5 then return "GetInfo"
	elseif opcode == 6 then return "Config"
	elseif opcode == 7 then return "GetConfig"
	elseif opcode == 8 then return "SetConfig"
	elseif opcode == 9 then return "SetPassword"
	elseif opcode == 10 then return "Report"
	elseif opcode == 11 then return "GetReport"
	elseif opcode == 12 then return "SetPreset"
	elseif opcode == 13 then return "SetUpdate"
	elseif opcode == 14 then return "UpdateProgress"
	end
	
	return "Unknown"
end


function IsDeviceToHost(opcode)
	return opcode == 1 or opcode == 4 or opcode == 6 or opcode == 10 or opcode == 14
end


-- Returns the opcode
function m.Parse(proto, buffer, tree)
	tree = tree:add(proto, buffer(0, 16), "GS Header")

	tree:add(m.headerMagic, buffer(0, 8))

	local bytes = buffer(8, 2)
	local payloadSize = bit32.band(bytes:le_uint(), 0x7fff) - 16
	local t = tree:add_le(g_f.size16, bytes):append_text(" (Payload "..payloadSize.." bytes)")
	t:add_le(m.packetSize, bytes)
	t:add_le(m.isMultipacket, bytes)

	bytes = buffer(10, 1)
	local opcode = bytes:uint()
	local opName = " (" .. GetName(opcode) .. ")"
	tree:add(g_f.opcode, bytes):append_text(opName)
	tree:append_text(opName)

	bytes = buffer(11, 1)
	t = tree:add(g_f.flags8, bytes)
	t:add(m.isExclusive, bytes)

	local devToHost = IsDeviceToHost(opcode)

	if devToHost then
		t:add(m.isPasswordValid, bytes)
		t:add(m.isAccessGranted, bytes)
	else
		t:add(m.isNoMeter, bytes)
	end

	if devToHost then
		tree:add(m.sequence, buffer(12, 1))
		tree:add(m.devId, buffer(13, 3))
	else
		tree:add_le(m.controllerId, buffer(12, 4))
	end

	return opcode, payloadSize
end

return m
