local helpers = require("Helpers")
local tracker = require("Tracker")

local m =
{
	firmwareVersion = ProtoField.uint16("gs.info.firmwareVersion", "Firmware version", base.HEX),
	statusPort = ProtoField.uint16("gs.info.statusPort", "Status port", base.DEC),
	statusIP = ProtoField.ipv4("gs.info.statusIP", "Status IP"),
	pid = ProtoField.uint16("gs.info.pid", "Product ID", base.DEC),
	variantId = ProtoField.uint8("gs.info.variantId", "Variant ID", base.DEC),
	protocolVersion = ProtoField.uint8("gs.info.protocolVersion", "Protocol version", base.DEC),
	dntUpdateable = ProtoField.bool("gs.info.dntUpdateable", "DNT updateable", 16, nil, 0x0001),
	dfuUpdateable = ProtoField.bool("gs.info.dfuUpdateable", "DFU updateable", 16, nil, 0x0002),
	dfu2Updateable = ProtoField.bool("gs.info.dfu2Updateable", "DFU2 updateable", 16, nil, 0x0004),
	danteProcessId = ProtoField.uint16("gs.info.danteProcessId", "Dante process ID", base.HEX),
	danteId = ProtoField.uint64("gs.info.danteId", "Dante ID", base.HEX),
	hostName = ProtoField.stringz("gs.info.hostName", "Host name", base.ASCII),
	friendlyName = ProtoField.stringz("gs.info.friendlyName", "Friendly name", base.ASCII),
	domainName = ProtoField.stringz("gs.info.domainName", "Domain name", base.ASCII),
	dntVersion = ProtoField.uint16("gs.info.dntVersion", "DNT version", base.HEX),
	dfu2Version = ProtoField.uint16("gs.info.dfu2Version", "DFU2 version", base.HEX)
}

local fieldPid


function m.GetFields()
	local fields =
	{
		firmwareVersion = m.firmwareVersion,
		info_statusPort = m.statusPort,
		info_statusIP = m.statusIP,
		pid = m.pid,
		variantId = m.variantId,
		protocolVersion = m.protocolVersion,
		dntUpdateable = m.dntUpdateable,
		dfuUpdateable = m.dfuUpdateable,
		dfu2Updateable = m.dfu2Updateable,
		danteProcessId = m.danteProcessId,
		danteId = m.danteId,
		hostName = m.hostName,
		friendlyName = m.friendlyName,
		domainName = m.domainName,
		dntVersion = m.dntVersion,
		dfu2Version = m.dfu2Version
	}
	
	return fields
end


function m.OnFieldsRegistered()
	-- Must be done only after the corresponding ProtoFields were registered.
	fieldPid = Field.new("gs.info.pid")
end


function m.Parse(proto, buffer, ip, tree)
	tree = tree:add(proto, buffer(), "Info")
	
	local size = buffer:len()

	if size ~= 120 and size ~= 128 then
		helpers.ErrorSize(tree)
		return
	end
	
	local isLegacy = (size == 120)

	tree:add_le(m.firmwareVersion, buffer(0,2))
	tree:add(m.statusPort, buffer(2,2))
	tree:add(m.statusIP, buffer(4,4))
	local t = tree:add_le(m.pid, buffer(8,2))
	
	-- Record an association of this device's IP address with its PID
	local pid = fieldPid()()
	tracker.AddPid(ip, pid)
	t:append_text(" (".. tracker.GetDeviceName(pid) ..")")
	
	tree:add_le(m.variantId, buffer(10,1))
	tree:add_le(m.protocolVersion, buffer(11,1))
	
	if not isLegacy then
		local bytes = buffer(12,2)
		local subtree = tree:add_le(g_f.flags16, bytes)
		subtree:add_le(m.dntUpdateable, bytes)
		subtree:add_le(m.dfuUpdateable, bytes)
		subtree:add_le(m.dfu2Updateable, bytes)
	end
	
	tree:add_le(m.danteProcessId, buffer(14,2))
	tree:add(m.danteId, buffer(16,8))
	tree:add(m.hostName, buffer(24,32))
	tree:add(m.friendlyName, buffer(56,32))
	tree:add(m.domainName, buffer(88,32))
	
	if not isLegacy then
		tree:add_le(m.dntVersion, buffer(120,2))
		tree:add_le(m.dfu2Version, buffer(122,2))
	end
end

return m
