local helpers = require("Helpers")
local tracker = require("Tracker")
local device = require("Device")
local channelNames = require("ChannelNames")
local divine = require("Divine")
local spark = require("Spark")
local dark1616m = require("Dark1616M")

local m =
{
	queryType = ProtoField.uint8("gs.report.queryType", "Query type", base.DEC),
}

local fieldQueryType


function m.GetFields()
	local fields =
	{
		queryType = m.queryType
	}

	return fields
end


function m.OnFieldsRegistered()
	-- Must be done only after the corresponding ProtoFields were registered.
	fieldQueryType = Field.new("gs.report.queryType")
end


local function GetName(reportType)
	if reportType == 0 then return "Preset"
	elseif reportType == 1 then return "Device"
	elseif reportType == 2 then return "In Channel Names"
	elseif reportType == 3 then return "Out Channel Names"
	end
	
	return ""
end


function m.ParseQuery(proto, buffer, tree)
	tree = tree:add(proto, buffer(), "Query")
	
	if buffer:len() < 4 then
		helpers.ErrorSize(tree)
		return
	end
	
	local t = tree:add(m.queryType, buffer(0, 1))
	local queryType = fieldQueryType()()
	
	if queryType < 4 then
		t:append_text(" ("..GetName(queryType)..")")
		tree:append_text(" ("..GetName(queryType)..")")
	else
		tree:append_text(" ("..queryType..")")
	end
end


function m.ParseReport(proto, buffer, ip, tree)
	tree = tree:add(proto, buffer(), "Report")

	local pid = tracker.GetPid(ip)
	local size = buffer:len()

	-- A report may contain multiple blocks of different type
	
	while size > 0 do
		-- There must be a block header and a minimum payload of 4 bytes
		if size < 8 then
			helpers.ErrorSize(tree)
			return
		end

		local subTree = tree:add(proto, buffer(0, 4), "Block Header")

		-- NOTE: Apparently Fields can't be reused in a loop, so we don't use them here
		local bytes = buffer(0, 1)
		local t = subTree:add(g_f.type8, bytes)
		local reportType = bytes:uint()
		
		if reportType < 4 then
			local name = " ("..GetName(reportType)..")"
			tree:append_text(name)
			subTree:append_text(name)
			t:append_text(name)
		elseif pid ~= 0 then
			local name = " ("..tracker.GetDeviceName(pid).." "..reportType..")"
			tree:append_text(name)
			subTree:append_text(name)
			t:append_text(name)
		else
			tree:append_text(" (type "..reportType..")")
		end
		
		subTree:add(g_f.generation, buffer(1, 1))
		bytes = buffer(2, 2)
		t = subTree:add_le(g_f.length16, bytes)
		local reportSize = bytes:le_uint() * 4
		t:append_text(" ("..reportSize.." bytes)")

		buffer = buffer(4)				-- skip block header
		local consumed = buffer:len()	-- in case not parsed
		
		if reportType == 1 then
			consumed = device.ParseReport(proto, buffer, reportSize, tree)
		elseif reportType == 2 then
			consumed = channelNames.ParseReportIn(proto, buffer, reportSize, tree)
		elseif reportType == 3 then
			consumed = channelNames.ParseReportOut(proto, buffer, reportSize, tree)	
		else
			if pid == 0 then
				tree:append_text(" (PID unknown)")
				
				if dark1616m.MaybeValidReport(buffer, reportSize, reportType) then pid = 6
				elseif spark.MaybeValidReport(buffer, reportSize, reportType) then pid = 30
				elseif divine.MaybeValidReport(buffer, reportSize, reportType) then pid = 49
				end
			end
			
			if pid == 6 then
				consumed = dark1616m.ParseReport(proto, buffer, reportSize, reportType, tree)
			elseif pid == 30 then
				consumed = spark.ParseReport(proto, buffer, reportSize, reportType, tree)
			elseif pid == 49 then
				consumed = divine.ParseReport(proto, buffer, reportSize, reportType, tree)
			end
		end

		if consumed < buffer:len() then
			buffer = buffer(consumed)
			size = buffer:len()
		else
			size = 0
		end
	end
end


return m

